#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Optional
from typing import Union

import numpy as np

from disent.schedule.lerp import cyclical_anneal
from disent.schedule.lerp import lerp
from disent.schedule.lerp import lerp_step

# ========================================================================= #
# Schedules                                                                 #
# ========================================================================= #


class Schedule(object):
    def __call__(self, step: int, value):
        return self.compute_value(step=step, value=value)

    def compute_value(self, step: int, value):
        raise NotImplementedError


# ========================================================================= #
# Value Schedules                                                           #
# ========================================================================= #


class NoopSchedule(Schedule):
    def compute_value(self, step: int, value):
        # does absolutely nothing!
        return value


class MultiplySchedule(Schedule):
    """
    A schedule that always applies a constant multiplier/ratio to the input value
    """

    # This schedule will always return a constant value!

    def __init__(self, r: float = 1.0):
        """
        :param r: The constant ratio of the original value that the schedule will use
        """
        self.r = r

    def compute_value(self, step: int, value):
        # we always return a constant value!
        return value * self.r


class FixedValueSchedule(Schedule):
    """
    Set a new constant value, instead of using the value passed to `compute_value`.
    - We can override config values using this class
    """

    def __init__(
        self,
        value: float,
        schedule: Optional[Schedule] = None,
    ):
        """
        :param schedule: The wrapped schedule that is passed the new constant value
        :param value: The value that should be used to replace the original values from the config.
                      If `compute_value` is called, the value passed to the function is replaced with this one!
        """
        assert (schedule is None) or isinstance(schedule, Schedule)
        self.schedule = schedule
        self.value = value

    def compute_value(self, step: int, value):
        del value
        # we override the passed value, and pass in a constant value instead!
        if self.schedule is None:
            return self.value
        else:
            return self.schedule(step, self.value)


# ========================================================================= #
# Value Schedules                                                           #
# ========================================================================= #


def _common_lerp_value(ratio, value, r_start: float, r_end: float):
    # scale the value such that it (which should be in the range [0, 1]) between [r_min, r_max]
    return lerp(
        ratio,
        start_val=value * r_start,
        end_val=value * r_end,
    )


def _completion_ratio(step: int, start_step: int, end_step: int):
    ratio = lerp_step(
        step=(step - start_step),
        max_step=(end_step - start_step),
        start_val=0.0,
        end_val=1.0,
    )
    return ratio


class LinearSchedule(Schedule):
    """
    A simple lerp schedule based on some start and end ratio.

    Multiples the value based on the step by some
    computed value that is in the range [0, 1]
    """

    def __init__(
        self,
        start_step: int,
        end_step: int,
        r_start: float = 0.0,
        r_end: float = 1.0,
    ):
        """
        :param min_step: The step at which the schedule starts and the value unfreezes
        :param max_step: The step at which the schedule finishes and the value freezes
        :param r_start: The ratio of the original value that the schedule will start with
        :param r_end: The ratio of the original value that the schedule will end with
        """
        assert start_step >= 0
        assert end_step > 0
        assert start_step < end_step
        self.start_step = start_step
        self.end_step = end_step
        self.r_start = r_start
        self.r_end = r_end

    def compute_value(self, step: int, value):
        # completion ratio in range [0, 1]. If step < start_step return 0, if step > end_step return 1
        ratio = _completion_ratio(step=step, start_step=self.start_step, end_step=self.end_step)
        # lerp the value into the range [r_start * value, r_end * value] according to the ratio
        return _common_lerp_value(ratio, value=value, r_start=self.r_start, r_end=self.r_end)


class CyclicSchedule(Schedule):
    """
    Cyclical schedule based on:
    https://arxiv.org/abs/1903.10145

    Multiples the value based on the step by some
    computed value that is in the range [0, 1]
    """

    # TODO: maybe move this api into cyclical_anneal
    def __init__(
        self,
        period: int,
        start_step: Optional[int] = None,
        repeats: Optional[int] = None,
        r_start: float = 0.0,
        r_end: float = 1.0,
        end_mode: str = "end",
        mode: str = "linear",
        p_low: float = 0.0,
        p_high: float = 0.0,
    ):
        """
        :param period: The number of steps it takes for the schedule to repeat
        :param start_step: The step when the schedule will start, if this is None
                           then no modification to the step is performed. Equivalent to
                           `start_step=0` if no negative step values are passed.
        :param repeats: The number of repeats of this schedule. The end_step of the schedule will
                        be `start_step + repeats*period`. If `repeats is None` or `repeats < 0` then the
                        schedule never ends.
        :param r_start: The ratio of the original value that the schedule will start with
        :param r_end: The ratio of the original value that the schedule will end with
        :param end_mode: what of value the schedule should take after finishing [start, end]
        :param mode: The kind of function use to interpolate between the start and finish [linear, sigmoid, cosine]
        :param p_low: The portion of the period at the start that is spent at the minimum value
        :param p_high: The portion of the period that at the end is spent at the maximum value
        """
        # checks
        if (repeats is not None) and (repeats < 0):
            repeats = None
        # set values
        self.period = period
        self.repeats = repeats
        self.start_step = start_step
        self.end_value = {"start": "low", "end": "high"}[end_mode]
        self.mode = mode
        # scale values
        self.r_start = r_start
        self.r_end = r_end
        # portions of low and high -- low + high <= 1.0 -- low + slope + high == 1.0
        self.p_low = p_low
        self.p_high = p_high
        # checks
        assert (start_step is None) or (start_step >= 0)

    def compute_value(self, step: int, value):
        # shift the start
        if self.start_step is not None:
            step = max(0, step - self.start_step)
        # outputs value in range [0, 1]
        ratio = cyclical_anneal(
            step=step,
            period=self.period,
            low_ratio=self.p_low,
            high_ratio=self.p_high,
            repeats=self.repeats,
            start_low=True,
            end_value=self.end_value,
            mode=self.mode,
        )
        return _common_lerp_value(ratio, value=value, r_start=self.r_start, r_end=self.r_end)


class SingleSchedule(CyclicSchedule):
    """
    A single repeat version of CyclicSchedule that automatically
    chooses if its going from high to low or low to high based on the start_value.

    Multiples the value based on the step by some
    computed value that is in the range [0, 1]
    """

    def __init__(
        self,
        start_step: int,
        end_step: int,
        r_start: float = 0.0,
        r_end: float = 1.0,
        mode: str = "linear",
    ):
        """
        :param start_step: The step when the schedule will start
        :param end_step: The step when the schedule will finish
        :param r_start: The ratio of the original value that the schedule will start with
        :param r_end: The ratio of the original value that the schedule will end with
        :param mode: The kind of function use to interpolate between the start and finish [linear, sigmoid, cosine]
        """
        super().__init__(
            period=(end_step - start_step),
            start_step=start_step,
            repeats=1,
            r_start=r_start,
            r_end=r_end,
            end_mode="end",
            mode=mode,
            p_low=0.0,  # adjust the start and end steps instead
            p_high=0.0,  # adjust the start and end steps instead
        )


class CosineWaveSchedule(Schedule):
    """
    A simple cosine wave schedule based on some start and end ratio.
    -- note this starts at zero by default

    Multiples the value based on the step by some
    computed value that is in the range [0, 1]
    """

    # TODO: add r_start
    # TODO: add start_step
    # TODO: add repeats
    def __init__(
        self,
        period: int,
        r_start: float = 0.0,
        r_end: float = 1.0,
    ):
        """
        :param period: The number of steps it takes for the schedule to repeat
        :param r_start: The ratio of the original value that the schedule will start with
        :param r_end: The ratio of the original value that the schedule will end with
        """
        assert period > 0
        self.period = period
        self.r_start = r_start
        self.r_end = r_end

    def compute_value(self, step: int, value):
        cosine_ratio = 0.5 * (1 + np.cos(step * (2 * np.pi / self.period) + np.pi))
        # lerp the value into the range [r_start * value, r_end * value] according to the ratio
        return _common_lerp_value(cosine_ratio, value=value, r_start=self.r_start, r_end=self.r_end)


# ========================================================================= #
# Clip Schedules                                                            #
# ========================================================================= #


class ClipSchedule(Schedule):
    """
    This schedule shifts the step, or clips the value
    """

    def __init__(
        self,
        schedule: Schedule,
        min_step: Optional[int] = None,
        max_step: Optional[int] = None,
        shift_step: Union[bool, int] = True,
        min_value: Optional[float] = None,
        max_value: Optional[float] = None,
    ):
        """
        :param schedule:
        :param min_step: The minimum step passed to the sub-schedule
        :param max_step: The maximum step passed to the sub-schedule
        :param shift_step: (if bool) Shift all the step values passed to the sub-schedule,
                           at or before min_step the sub-schedule will get `0`, at or after
                           max_step the sub-schedule will get `max_step-shift_step`
                           (if int) Add the given value to the step passed to the sub-schedule
        :param min_value: The minimum value returned from the sub-schedule
        :param max_value: The maximum value returned from the sub-schedule
        """
        assert isinstance(schedule, Schedule)
        self.schedule = schedule
        # step settings
        self.min_step = min_step
        self.max_step = max_step
        # shift step
        self.shift_step = shift_step
        if isinstance(shift_step, bool):
            if self.min_step is not None:
                self.shift_step = -self.min_step
        # value settings
        self.min_value = min_value
        self.max_value = max_value

    def compute_value(self, step: int, value):
        if self.max_step is not None:
            step = np.minimum(self.max_step, step)
        if self.min_step is not None:
            step = np.maximum(self.min_step, step)
        if self.shift_step is not None:
            step += self.shift_step
        result = self.schedule(step, value)
        if self.max_value is not None:
            result = np.minimum(self.max_value, result)
        if self.min_value is not None:
            result = np.maximum(self.min_value, result)
        return result


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


if __name__ == "__main__":

    def plot_schedules(*schedules: Schedule, total: int = 1000, value=1):
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(1, len(schedules), figsize=(3 * len(schedules), 3))
        for ax, s in zip(axs, schedules):
            xs = list(range(total))
            ys = [s(x, value) for x in xs]
            ax.set_xlim([-0.05 * total, total + 0.05 * total])
            ax.set_ylim([-0.05 * value, value + 0.05 * value])
            ax.plot(xs, ys)
            ax.set_title(f"{s.__class__.__name__}")
        fig.tight_layout()
        plt.show()

    def main():
        # these should be equivalent
        plot_schedules(
            LinearSchedule(start_step=100, end_step=900, r_start=0.1, r_end=0.8),
            LinearSchedule(start_step=200, end_step=800, r_start=0.9, r_end=0.2),
            SingleSchedule(start_step=100, end_step=900, r_start=0.1, r_end=0.8),
            SingleSchedule(start_step=200, end_step=800, r_start=0.9, r_end=0.2),
            # LinearSchedule(min_step=900, max_step=100, r_start=0.1, r_end=0.8), # INVALID
            # LinearSchedule(min_step=900, max_step=100, r_start=0.8, r_end=0.1), # INVALID
        )

        plot_schedules(
            CyclicSchedule(
                period=300,
                start_step=0,
                repeats=None,
                r_start=0.1,
                r_end=0.8,
                end_mode="end",
                mode="linear",
                p_low=0.00,
                p_high=0.00,
            ),
            CyclicSchedule(
                period=300,
                start_step=0,
                repeats=2,
                r_start=0.9,
                r_end=0.2,
                end_mode="start",
                mode="linear",
                p_low=0.25,
                p_high=0.00,
            ),
            CyclicSchedule(
                period=300,
                start_step=200,
                repeats=2,
                r_start=0.9,
                r_end=0.2,
                end_mode="end",
                mode="linear",
                p_low=0.00,
                p_high=0.25,
            ),
            CyclicSchedule(
                period=300,
                start_step=0,
                repeats=2,
                r_start=0.1,
                r_end=0.8,
                end_mode="end",
                mode="cosine",
                p_low=0.25,
                p_high=0.25,
            ),
            CyclicSchedule(
                period=300,
                start_step=250,
                repeats=None,
                r_start=0.1,
                r_end=0.8,
                end_mode="end",
                mode="sigmoid",
                p_low=0.00,
                p_high=0.00,
            ),
        )

        plot_schedules(
            SingleSchedule(start_step=0, end_step=800, r_start=0.1, r_end=0.8, mode="linear"),
            SingleSchedule(start_step=100, end_step=800, r_start=0.8, r_end=0.1, mode="linear"),
            SingleSchedule(start_step=100, end_step=800, r_start=0.8, r_end=0.1, mode="linear"),
        )

    main()



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import inspect
import re
from abc import ABC
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Iterator
from typing import List
from typing import MutableMapping
from typing import NoReturn
from typing import Optional
from typing import Protocol
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union

from disent.util.function import wrapped_partial
from disent.util.imports import _check_and_split_path
from disent.util.imports import import_obj
from disent.util.imports import import_obj_partial

# ========================================================================= #
# Type Hints                                                                #
# ========================================================================= #


K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")
AliasesHint = Union[str, Tuple[str, ...]]


class _FactoryFn(Protocol[V]):
    def __call__(self, *args) -> V:
        ...


# ========================================================================= #
# Provided Values                                                           #
# ========================================================================= #


class ProvidedValue(Generic[V], ABC):
    """
    Base class for providing immutable values using the `get` method.
    - Subclasses should override this
    """

    def get(self) -> V:
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class StaticValue(ProvidedValue[V]):
    """
    Provide static values. Simply a see-through wrapper
    around already generated / constant values.
    """

    def __init__(self, value: V):
        self._value = value

    def get(self) -> V:
        return self._value

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._value)})"


class StaticImport(StaticValue[V]):
    def __init__(self, fn: V, *partial_args, **partial_kwargs):
        super().__init__(wrapped_partial(fn, *partial_args, **partial_kwargs))


class LazyValue(ProvidedValue[V]):
    """
    Use a function to provide a value by generating and caching
    the result only when this value is first needed.
    """

    def __init__(self, generate_fn: Callable[[], V]):
        assert callable(generate_fn)
        self._generate_fn = generate_fn
        self._is_generated = False
        self._value = None

    def get(self) -> V:
        # cache the value
        if not self._is_generated:
            self._is_generated = True
            self._value = self._generate_fn()
        return self._value

    def clear(self):
        self._is_generated = False
        self._value = None

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._generate_fn)})"


class LazyImport(LazyValue[V]):
    """
    Like lazy value, but instead takes in the import path to a callable object.
    Any remaining args and kwargs are used to partially parameterize the object.
    - The partial object is returned, and not called. This should be
      the same as importing the value directly when `get` is called!
    """

    def __init__(self, import_path: str, *partial_args, **partial_kwargs):
        # function imports the object when called
        def generate_fn():
            return import_obj_partial(import_path, *partial_args, **partial_kwargs)

        # initialise the lazy value
        super().__init__(generate_fn=generate_fn)


# ========================================================================= #
# Provided Dict                                                             #
# ========================================================================= #


class DictProviders(MutableMapping[K, V]):
    """
    A dictionary that only allows instances of
    provided values to be added to it.
    - The returned values are obtained directly from the providers
    """

    def __init__(self):
        self._providers: Dict[K, ProvidedValue[V]] = {}

    def __getitem__(self, k: K) -> V:
        return self._getitem(k)

    def __contains__(self, k: K):
        return k in self._providers

    def __setitem__(self, k: K, v: ProvidedValue[V]) -> NoReturn:
        self._setitem(k, v)

    def __delitem__(self, k: K) -> NoReturn:
        del self._providers[k]

    def __len__(self) -> int:
        return len(self._providers)

    def __iter__(self) -> Iterator[K]:
        yield from self._providers

    # allow easier overriding in subclasses without calling super() which can get confusing

    def _getitem(self, k: K) -> V:
        provider = self._providers[k]
        return provider.get()

    def _setitem(self, k: K, v: ProvidedValue[V]) -> NoReturn:
        if not isinstance(v, ProvidedValue):
            raise TypeError(
                f"Values stored in {self.__class__.__name__} must be instances of: {ProvidedValue.__name__}, got: {repr(v)}"
            )
        self._providers[k] = v


# ========================================================================= #
# Registry - Mixin                                                          #
# ========================================================================= #


class Registry(DictProviders[str, V]):
    def __init__(self, name: str):
        if not str.isidentifier(name):
            raise ValueError(f"Registry names must be valid identifiers, got: {repr(name)}")
        # initialise
        self._name = name
        super().__init__()

    @property
    def static_examples(self) -> List[str]:
        return list(self._providers.keys())

    @property
    def examples(self) -> List[str]:
        return self.static_examples

    @property
    def name(self) -> str:
        return self._name

    def __repr__(self):
        return f"{self.__class__.__name__}({self._name})"

    # --- CORE --- #

    def __setitem__(self, aliases: AliasesHint, v: ProvidedValue[V]) -> NoReturn:
        self._setitems(aliases, v)

    def __getitem__(self, k: str) -> V:
        value = self._getitem(k)
        self._check_provided_value(value)
        return value

    def __delitem__(self, k: str) -> None:
        raise RuntimeError(
            f"Registry: {repr(self.name)} does not support item deletion. Tried to remove key: {repr(k)}"
        )

    # --- HELPER --- #

    def _setitems(self, aliases: AliasesHint, v: Union[V, ProvidedValue[V]]) -> None:
        aliases = self._normalise_aliases(aliases)
        # check all the aliases
        for k in aliases:
            if not str.isidentifier(k):
                raise ValueError(
                    f"Keys stored in registry: {repr(self.name)} must be valid identifiers, got: {repr(k)}"
                )
            if k in self:
                raise RuntimeError(f"Tried to overwrite existing key: {repr(k)} in registry: {repr(self.name)}")
            self._check_key(k)
        # check the value
        v = self._check_and_normalise_value(v)
        # set all the aliases
        for k in aliases:
            self._setitem(k, v)

    def _normalise_aliases(self, aliases: AliasesHint, check_nonempty: bool = True) -> Tuple[str]:
        if isinstance(aliases, str):
            aliases = (aliases,)
        if not isinstance(aliases, tuple):
            raise TypeError(
                f"Multiple aliases must be provided to registry: {repr(self.name)} as a Tuple[str], got: {repr(aliases)}"
            )
        if check_nonempty:
            if len(aliases) < 1:
                raise ValueError(
                    f"At least one alias must be provided to registry: {repr(self.name)}, got: {repr(aliases)}"
                )
        return aliases

    # --- OVERRIDABLE --- #

    def _check_and_normalise_value(self, v: ProvidedValue[V]) -> ProvidedValue[V]:
        return v

    def _check_provided_value(self, v: V) -> NoReturn:
        pass

    def _check_key(self, k: str) -> NoReturn:
        pass

    # --- MISSING VALUES --- #

    def setmissing(self, alias: AliasesHint, value: V) -> NoReturn:
        # find missing keys
        aliases = self._normalise_aliases(alias)
        missing = tuple(alias for alias in aliases if (alias not in self))
        # register missing keys
        if missing:
            self._setitems(missing, value)

    @property
    def setm(self) -> "_RegistrySetMissing":
        # instead of checking values manually, at the cost of some efficiency,
        # this allows us to register values multiple times with hardly modified notation!
        # -- only modifies unset values
        # set once:    `REGISTRY['key'] = val`
        # set default: `REGISTRY.setm['key'] = val`
        return self._RegistrySetMissing(self)

    class _RegistrySetMissing(object):
        def __init__(self, registry: "Registry"):
            self._registry = registry

        def __setitem__(self, aliases: str, v: ProvidedValue[V]) -> NoReturn:
            self._registry.setmissing(aliases, v)


# ========================================================================= #
# Import Registry                                                           #
# ========================================================================= #


# TODO: merge this with the dynamic registry below?
class RegistryImports(Registry[V]):
    """
    A registry for arbitrary imports.
    -- supports decorating functions and classes
    """

    def register(
        self,
        aliases: Optional[AliasesHint] = None,
        auto_alias: bool = True,
        partial_args: Tuple[Any, ...] = None,
        partial_kwargs: Dict[str, Any] = None,
    ) -> Callable[[T], T]:
        """
        Register a function or object to this registry.
        - can be used as a decorator @register(...)
        - automatically chooses an alias based on the function name
        - specify defaults for the function with the args and kwargs
        """
        # default values
        if aliases is None:
            aliases = ()
        if partial_args is None:
            partial_args = ()
        if partial_kwargs is None:
            partial_kwargs = {}
        aliases = self._normalise_aliases(aliases, check_nonempty=False)

        # add the function name as an alias if it does not already exist,
        # then register the partially parameterised function as a static value
        def _decorator(orig_fn):
            keys = self._append_auto_alias(self._get_fn_alias(orig_fn), aliases=aliases, auto_alias=auto_alias)
            self[keys] = StaticImport(orig_fn, *partial_args, **partial_kwargs)
            return orig_fn

        return _decorator

    def register_import(
        self,
        import_path: str,
        aliases: Optional[AliasesHint] = None,
        auto_alias: bool = True,
        *partial_args,
        **partial_kwargs,
    ) -> NoReturn:
        """
        Register an import path and automatically obtain an alias from it.
        - This is the same as: registry[(import_name, *aliases)] = LazyImport(import_path, *partial_args, **partial_kwargs)
        """
        # normalise aliases
        if aliases is None:
            aliases = ()
        aliases = self._normalise_aliases(aliases, check_nonempty=False)
        # add object alias
        (*_, alias) = _check_and_split_path(import_path)
        aliases = self._append_auto_alias(alias, aliases=aliases, auto_alias=auto_alias)
        # register the lazy import
        self[aliases] = LazyImport(import_path=import_path, *partial_args, **partial_kwargs)

    # --- ALIAS HELPER --- #

    def _append_auto_alias(self, alias: Optional[str], aliases: Tuple[str, ...], auto_alias: bool):
        if auto_alias:
            if alias is not None:
                if alias not in self:
                    aliases = (alias, *aliases)
                elif not aliases:
                    raise RuntimeError(
                        f"automatic alias: {repr(alias)} already exists for registry: {repr(self.name)} and no alternative aliases were specified."
                    )
            elif not aliases:
                raise RuntimeError(f"Cannot add value to registry: {repr(self.name)}, no automatic alias was found!")
        elif not aliases:
            raise RuntimeError(
                f"Cannot add value to registry: {repr(self.name)}, no manual aliases were specified and automatic aliasing is disabled!"
            )
        return aliases

    @staticmethod
    def _get_fn_alias(fn) -> Optional[str]:
        if hasattr(fn, "__name__"):
            if str.isidentifier(fn.__name__):
                return fn.__name__
        return None

    # --- OVERRIDDEN --- #

    def _check_and_normalise_value(self, v: ProvidedValue[V]) -> ProvidedValue[V]:
        if not isinstance(v, (LazyImport, StaticImport)):
            raise TypeError(
                f"Values stored in registry: {repr(self.name)} must be instances of: {(LazyImport.__name__, StaticImport.__name__)}, got: {repr(v)}"
            )
        return v

    # --- OVERRIDABLE --- #

    def _check_provided_value(self, v: V) -> NoReturn:
        pass

    def _check_key(self, k: str) -> NoReturn:
        pass


# ========================================================================= #
# Dynamic Registry                                                          #
# ========================================================================= #

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# Constructor                                                               #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class RegexConstructor(object):
    def __init__(
        self,
        pattern: Union[str, re.Pattern],
        example: str,
        factory_fn: Union[_FactoryFn[V], str],
    ):
        self._pattern = self._check_pattern(pattern)
        self._example = self._check_example(example, self._pattern)
        # we can delay loading of the function if it is a string!
        self._factory_fn = (
            factory_fn if isinstance(factory_fn, str) else self._check_factory_fn(factory_fn, self._pattern)
        )

    @classmethod
    def _check_pattern(cls, pattern: Union[str, re.Pattern]):
        # check the regex type & convert
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        if not isinstance(pattern, re.Pattern):
            raise TypeError(f"regex pattern must be a regex `str` or `re.Pattern`, got: {repr(pattern)}")
        if pattern.groups < 1:
            raise ValueError(f"regex pattern must contain at least one group, got: {repr(pattern)}")
        return pattern

    @classmethod
    def _check_factory_fn(cls, factory_fn: _FactoryFn[V], pattern: re.Pattern) -> _FactoryFn[V]:
        # we have an actual function, we can check it!
        if not callable(factory_fn):
            raise TypeError(f"generator function must be callable, got: {factory_fn}")
        signature = inspect.signature(factory_fn)
        if len(signature.parameters) != pattern.groups:
            raise ValueError(
                f"signature has incorrect number of parameters: {repr(signature)} compared to the number of groups in the regex pattern: {repr(pattern)}"
            )
        return factory_fn

    @classmethod
    def _check_example(cls, example: str, pattern: re.Pattern) -> str:
        # check the example
        if not isinstance(example, str):
            raise TypeError(f"example must be a `str`, got: {type(example)}")
        if not example:
            raise ValueError(f"example must not be empty, got: {type(example)}")
        # check that the regex matches the example!
        if pattern.search(example) is None:
            raise ValueError(f"could not match example: {repr(example)} to regex: {repr(pattern)}")
        return example

    @property
    def pattern(self) -> re.Pattern:
        return self._pattern

    @property
    def example(self) -> str:
        return self._example

    def construct(self, name: str) -> V:
        # get the results
        result = self._pattern.search(name)
        if result is None:
            raise KeyError(
                f"pattern: {self.pattern} does not match given name: {repr(name)}. The following example would be valid: {repr(self.example)}"
            )
        # get the function -- load via the path
        if isinstance(self._factory_fn, str):
            fn = import_obj(self._factory_fn)
            self._factory_fn = self._check_factory_fn(fn, self._pattern)
        # construct
        return self._factory_fn(*result.groups())

    def can_construct(self, name: str) -> bool:
        return self._pattern.search(name) is not None


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# Cached Linear Search                                                      #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class RegexProvidersSearch(object):
    def __init__(self):
        self._patterns: Set[re.Pattern] = set()
        self._constructors: List[RegexConstructor] = []
        # caching
        self._cache = {}
        self._cache_dirty = False

    @property
    def regex_constructors(self) -> List[RegexConstructor]:
        return list(self._constructors)

    def construct(self, arg_str: str):
        provider = self.get_constructor(arg_str)
        # build the object
        if provider is not None:
            return provider.construct(arg_str)
        # no result was found!
        raise KeyError(
            f"could not construct an item from the given argument string: {repr(arg_str)}, valid patterns include: {[p.pattern for p in self._constructors]}"
        )

    def can_construct(self, arg_str: str) -> bool:
        return self.get_constructor(arg_str) is not None

    def get_constructor(self, arg_str: str) -> Optional[RegexConstructor]:
        # TODO: clean up this cache!
        # check cache -- remove None entries if dirty
        if self._cache_dirty:
            self._cache = {k: v for k, v in self._cache.items() if v is not None}
            self._cache_dirty = False
        if arg_str in self._cache:
            return self._cache[arg_str]
        # check the input string
        if not isinstance(arg_str, str):
            raise TypeError(f"regex factory can only construct from `str`, got: {repr(arg_str)}")
        if not arg_str:
            raise ValueError(f"regex factory can only construct from non-empty `str`, got: {repr(arg_str)}")
        # match the values
        constructor = None
        for c in self:
            if c.can_construct(arg_str):
                constructor = c
                break
        # cache the value
        self._cache[arg_str] = constructor
        if len(self._cache) > 128:
            self._cache.popitem()
        return constructor

    def has_pattern(self, pattern: Union[str, re.Pattern]) -> bool:
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        return pattern in self._patterns

    def __len__(self) -> int:
        return len(self._constructors)

    def __iter__(self) -> Iterator[RegexConstructor]:
        yield from self._constructors

    def append(self, constructor: RegexConstructor):
        if not isinstance(constructor, RegexConstructor):
            raise TypeError(f"regex factory only accepts {RegexConstructor.__name__} providers.")
        if constructor.pattern in self._patterns:
            raise RuntimeError(f"regex factory already contains the regex pattern: {repr(constructor.pattern)}")
        # append value!
        self._patterns.add(constructor.pattern)
        self._constructors.append(constructor)
        self._cache_dirty = True


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
# Dynamic Registry                                                          #
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #


class RegexRegistry(Registry[V]):

    """
    Registry that allows registering of regex expressions that can be used to
    construct values if there is no static value found!
    - Regular expressions are checked in the order they are registered.
    - `name in registry` checks if any of the expression matches, it does not check for an existing regex
    - `len(registry)` returns the number of examples available, each item & regex factory
    - `for example in registry` returns the examples available, each item & regex factory should be called if we use these to construct values `registry[example]`

    To check for an already added regex expression, use:
    - `has_regex(expr)`
    """

    def __init__(self, name: str):
        self._regex_providers = RegexProvidersSearch()
        super().__init__(name)

    # --- CORE ... UPDATED WITH LINEAR SEARCH --- #

    @property
    def regex_constructors(self) -> List[RegexConstructor]:
        return self._regex_providers.regex_constructors

    @property
    def regex_examples(self) -> List[str]:
        return [constructor.example for constructor in self._regex_providers.regex_constructors]

    @property
    def examples(self) -> List[str]:
        return [*self.static_examples, *self.regex_examples]

    def __getitem__(self, k: str) -> V:
        assert isinstance(k, str), f"invalid key: {repr(k)}, must be a `str`"
        # the regex provider is cached so this should be efficient for the same value calls
        # -- we do not cache the actual provided value!
        if k in self._providers:
            return self._getitem(k)
        elif self._regex_providers.can_construct(k):
            return self._regex_providers.construct(k)
        raise KeyError(
            f"dynamic registry: {repr(self.name)} cannot construct item with key: {repr(k)}. Valid static values: {sorted(self._providers.keys())}. Valid dynamic examples: {[p.example for p in self._regex_providers]}"
        )

    def __setitem__(self, aliases: AliasesHint, v: ProvidedValue[V]) -> NoReturn:
        if isinstance(aliases, re.Pattern) or isinstance(v, RegexConstructor):
            raise RuntimeError(
                f"register dynamic values to the dynamic registry: {repr(self.name)} with the `register_regex` or `register_constructor` methods."
            )
        super().__setitem__(aliases, v)

    def __contains__(self, k: K):
        if k in self._providers:
            return True
        if self._regex_providers.can_construct(k):
            return True
        return False

    def __len__(self) -> int:
        return len(self._providers) + len(self._regex_providers)

    def __iter__(self) -> Iterator[K]:
        yield from self._providers
        yield from (p.example for p in self._regex_providers)

    # --- OVERRIDABLE --- #

    def _check_regex_constructor(self, constructor: RegexConstructor):
        pass

    # --- DYNAMIC VALUES --- #

    def has_regex(self, pattern: Union[str, re.Pattern]) -> bool:
        return self._regex_providers.has_pattern(pattern)

    def register_constructor(self, constructor: RegexConstructor) -> "RegexRegistry":
        """
        Register a regex constructor
        """
        if not isinstance(constructor, RegexConstructor):
            raise TypeError(
                f"dynamic registry: {repr(self.name)} only accepts dynamic {RegexConstructor.__name__}, got: {repr(constructor)}"
            )
        self._check_regex_constructor(constructor)
        self._regex_providers.append(constructor)
        return self

    def register_regex(
        self, pattern: Union[str, re.Pattern], example: str, factory_fn: Optional[Union[_FactoryFn[V], str]] = None
    ):
        """
        Register and create a regex constructor
        """

        def _register_wrapper(fn: T) -> T:
            self.register_constructor(RegexConstructor(pattern=pattern, example=example, factory_fn=fn))
            return fn

        return _register_wrapper if (factory_fn is None) else _register_wrapper(factory_fn)

    def register_missing_constructor(self, constructor: RegexConstructor):
        """
        Only register a regex constructor if the pattern does not already exist!
        """
        if not self.has_regex(constructor.pattern):
            return self.register_constructor(constructor)

    def register_missing_regex(
        self, pattern: Union[str, re.Pattern], example: str, factory_fn: Optional[Union[_FactoryFn[V], str]] = None
    ):
        """
        Only register and create a regex constructor if the pattern does not already exist!
        """
        if not self.has_regex(pattern):
            return self.register_regex(pattern=pattern, example=example, factory_fn=factory_fn)
        elif factory_fn is None:
            return lambda fn: fn  # dummy wrapper

    # --- MISSING VALUES --- #

    # override from the parent class!
    class _RegistrySetMissing(Registry._RegistrySetMissing):
        _registry: "RegexRegistry"

        def register_constructor(self, constructor: RegexConstructor):
            """
            Only register a regex constructor if the pattern does not already exist!
            """
            return self._registry.register_missing_constructor(constructor=constructor)

        def register_regex(
            self, pattern: Union[str, re.Pattern], example: str, factory_fn: Optional[Union[_FactoryFn[V], str]] = None
        ):
            """
            Only register and create a regex constructor if the pattern does not already exist!
            """
            return self._registry.register_missing_regex(
                pattern=pattern,
                example=example,
                factory_fn=factory_fn,
            )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import pytest
import torch

from disent.nn.functional._util_generic import generic_as_int32
from disent.nn.functional._util_generic import generic_max
from disent.nn.functional._util_generic import generic_min
from disent.nn.functional._util_generic import generic_ndim
from disent.nn.functional._util_generic import generic_shape

# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _assert_type_and_value(input, target):
    # check types are the same
    assert type(input) == type(target)
    # specific checks
    if isinstance(target, (int, float)):
        assert input == target
    elif isinstance(target, (np.ndarray, np.int32, np.int64, np.float32, np.float64)):
        assert input.dtype == target.dtype
        assert np.all(input == target)
    elif isinstance(target, torch.Tensor):
        assert input.dtype == target.dtype
        assert input.device == target.device
        assert torch.all(input == target)
    else:  # pragma: no cover
        raise NotImplementedError("This should never happen in tests!")


# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


def test_generic_as_int32():
    # scalars
    _assert_type_and_value(input=generic_as_int32(-1), target=-1)
    _assert_type_and_value(input=generic_as_int32(-1.0), target=-1)
    _assert_type_and_value(input=generic_as_int32(1.5), target=1)
    # torch
    _assert_type_and_value(
        input=generic_as_int32(torch.as_tensor([-1.5, 0, 1.0])), target=torch.as_tensor([-1, 0, 1], dtype=torch.int32)
    )
    # numpy
    _assert_type_and_value(
        input=generic_as_int32(np.array([-1.5, 0, 1.0])), target=np.array([-1, 0, 1], dtype=np.int32)
    )
    # unsupported
    with pytest.raises(TypeError, match="invalid type"):
        generic_as_int32(None)


def test_generic_max():
    # scalars
    _assert_type_and_value(input=generic_max(-1), target=-1)
    _assert_type_and_value(input=generic_max(1.5), target=1.5)
    # torch
    _assert_type_and_value(input=generic_max(torch.as_tensor([-1, 0, 1])), target=torch.as_tensor(1, dtype=torch.int64))
    _assert_type_and_value(
        input=generic_max(torch.as_tensor([-1.0, 0.0, 1.0])), target=torch.as_tensor(1.0, dtype=torch.float32)
    )
    # numpy
    _assert_type_and_value(input=generic_max(np.array([-1, 0, 1])), target=np.int64(1))
    _assert_type_and_value(input=generic_max(np.array([-1.0, 0.0, 1.0])), target=np.float64(1.0))
    # unsupported
    with pytest.raises(TypeError, match="invalid type"):
        generic_max(None)


def test_generic_min():
    # scalars
    _assert_type_and_value(input=generic_min(-1), target=-1)
    _assert_type_and_value(input=generic_min(1.5), target=1.5)
    # torch
    _assert_type_and_value(
        input=generic_min(torch.as_tensor([-1, 0, 1])), target=torch.as_tensor(-1, dtype=torch.int64)
    )
    _assert_type_and_value(
        input=generic_min(torch.as_tensor([-1.0, 0.0, 1.0])), target=torch.as_tensor(-1.0, dtype=torch.float32)
    )
    # numpy
    _assert_type_and_value(input=generic_min(np.array([-1, 0, 1])), target=np.int64(-1))
    _assert_type_and_value(input=generic_min(np.array([-1.0, 0.0, 1.0])), target=np.float64(-1.0))
    # unsupported
    with pytest.raises(TypeError, match="invalid type"):
        generic_min(None)


def test_generic_shape():
    # scalars
    assert generic_shape(1.0) == ()
    assert generic_shape(-1) == ()
    # torch
    assert generic_shape(torch.as_tensor([-1, 0, 1])) == (3,)
    assert generic_shape(torch.as_tensor([-1.0, 0.0, 1.0])) == (3,)
    # numpy
    assert generic_shape(np.array([-1, 0, 1])) == (3,)
    assert generic_shape(np.array([-1.0, 0.0, 1.0])) == (3,)
    # unsupported
    with pytest.raises(TypeError, match="invalid type"):
        generic_shape(None)


def test_generic_ndim():
    # scalars
    assert generic_ndim(1.0) == 0
    assert generic_ndim(-1) == 0
    # torch
    assert generic_ndim(torch.as_tensor([-1, 0, 1])) == 1
    assert generic_ndim(torch.as_tensor([-1.0, 0.0, 1.0])) == 1
    # numpy
    assert generic_ndim(np.array([-1, 0, 1])) == 1
    assert generic_ndim(np.array([-1.0, 0.0, 1.0])) == 1
    # unsupported
    with pytest.raises(TypeError, match="invalid type"):
        generic_ndim(None)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import warnings
from typing import Optional
from typing import Sequence

import lightning as L

from disent import registry as R
from disent.dataset.data import GroundTruthData
from disent.util.lightning.callbacks._callbacks_base import BaseCallbackPeriodic
from disent.util.lightning.callbacks._helper import _get_dataset_and_ae_like
from disent.util.lightning.logger_util import log_metrics
from disent.util.lightning.logger_util import wb_log_reduced_summaries
from disent.util.profiling import Timer
from disent.util.strings import colors as c

log = logging.getLogger(__name__)


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _normalized_numeric_metrics(items: dict):
    results = {}
    for k, v in items.items():
        if isinstance(v, (float, int)):
            results[k] = v
        else:
            try:
                results[k] = float(v)
            except:
                log.warning(
                    f"SKIPPED: metric with key: {repr(k)}, result has invalid type: {type(v)} with value: {repr(v)}"
                )
    return results


# ========================================================================= #
# Metrics Callback                                                          #
# ========================================================================= #


class VaeMetricLoggingCallback(BaseCallbackPeriodic):
    def __init__(
        self,
        step_end_metrics: Optional[Sequence[str]] = None,
        train_end_metrics: Optional[Sequence[str]] = None,
        every_n_steps: Optional[int] = None,
        begin_first_step: bool = False,
    ):
        super().__init__(every_n_steps, begin_first_step)
        self.step_end_metrics = step_end_metrics if step_end_metrics else []
        self.train_end_metrics = train_end_metrics if train_end_metrics else []
        assert isinstance(self.step_end_metrics, list)
        assert isinstance(self.train_end_metrics, list)
        assert (
            self.step_end_metrics or self.train_end_metrics
        ), "No metrics given to step_end_metrics or train_end_metrics"

    def _compute_metrics_and_log(self, trainer: L.Trainer, pl_module: L.LightningModule, metrics: list, is_final=False):
        # get dataset and vae framework from trainer and module
        dataset, vae = _get_dataset_and_ae_like(trainer, pl_module, unwrap_groundtruth=True)
        # check if we need to skip
        # TODO: dataset needs to be able to handle wrapped datasets!
        if not dataset.is_ground_truth:
            warnings.warn(
                f"{dataset.__class__.__name__} is not an instance of {GroundTruthData.__name__}. Skipping callback: {self.__class__.__name__}!"
            )
            return
        # get padding amount
        pad = max(7 + len(k) for k in R.METRICS)  # I know this is a magic variable... im just OCD
        # compute all metrics
        for metric in metrics:
            if is_final:
                log.info(f"| {metric.__name__:<{pad}} - computing...")
            with Timer() as timer:
                scores = metric(dataset, lambda x: vae.encode(x.to(vae.device)))
            metric_results = " ".join(f"{k}{c.GRY}={c.lMGT}{v:.3f}{c.RST}" for k, v in scores.items())
            log.info(f"| {metric.__name__:<{pad}} - time{c.GRY}={c.lYLW}{timer.pretty:<9}{c.RST} - {metric_results}")

            # log to trainer
            prefix = "final_metric" if is_final else "epoch_metric"
            prefixed_scores = {f"{prefix}/{k}": v for k, v in scores.items()}
            log_metrics(trainer.loggers, _normalized_numeric_metrics(prefixed_scores))

            # log summary for WANDB
            # this is kinda hacky... the above should work for parallel coordinate plots
            wb_log_reduced_summaries(trainer.loggers, prefixed_scores, reduction="max")

    def do_step(self, trainer: L.Trainer, pl_module: L.LightningModule):
        if self.step_end_metrics:
            log.debug("Computing Epoch Metrics:")
            with Timer() as timer:
                self._compute_metrics_and_log(trainer, pl_module, metrics=self.step_end_metrics, is_final=False)
            log.debug(f"Computed Epoch Metrics! {timer.pretty}")

    def on_train_end(self, trainer: L.Trainer, pl_module: L.LightningModule):
        if self.train_end_metrics:
            log.debug("Computing Final Metrics...")
            with Timer() as timer:
                self._compute_metrics_and_log(trainer, pl_module, metrics=self.train_end_metrics, is_final=True)
            log.debug(f"Computed Final Metrics! {timer.pretty}")


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple

import torch

from disent.frameworks.ae._unsupervised__ae import Ae
from disent.frameworks.vae._weaklysupervised__adavae import AdaVae

# ========================================================================= #
# Ada-GVAE                                                                  #
# ========================================================================= #


class AdaAe(Ae):
    """
    Custom implementation, removing Variational Auto-Encoder components of:
    Weakly Supervised Disentanglement Learning Without Compromises: https://arxiv.org/abs/2002.02886

    MODIFICATION:
    - L1 distance for deltas instead of KL divergence
    - adjustable threshold value
    """

    REQUIRED_OBS = 2

    @dataclass
    class cfg(Ae.cfg):
        ada_thresh_ratio: float = 0.5

    def hook_ae_intercept_zs(self, zs: Sequence[torch.Tensor]) -> Tuple[Sequence[torch.Tensor], Dict[str, Any]]:
        """
        Adaptive VAE Glue Method, putting the various components together
        1. find differences between deltas
        2. estimate a threshold for differences
        3. compute a shared mask from this threshold
        4. average together elements that should be considered shared

        TODO: the methods used in this function should probably be moved here
        TODO: this function could be turned into a torch.nn.Module!
        """
        z0, z1 = zs
        # shared elements that need to be averaged, computed per pair in the batch.
        share_mask = AdaVae.compute_shared_mask_from_zs(z0, z1, ratio=self.cfg.ada_thresh_ratio)
        # compute average posteriors
        new_zs = AdaVae.make_shared_zs(z0, z1, share_mask)
        # return new args & generate logs
        return new_zs, {"shared": share_mask.sum(dim=1).float().mean()}


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from itertools import islice
from typing import List
from typing import Sequence

# ========================================================================= #
# Iterators                                                                 #
# ========================================================================= #


def chunked(arr, chunk_size: int, include_remainder=True):
    """
    return an array of array chucks of size chunk_size.
    This is NOT an iterable, and returns all the data.
    """
    size = (len(arr) + chunk_size - 1) if include_remainder else len(arr)
    return [arr[chunk_size * i : chunk_size * (i + 1)] for i in range(size // chunk_size)]


def iter_chunks(items, chunk_size: int, include_remainder=True):
    """
    iterable version of chunked.
    that does not evaluate unneeded elements
    """
    items = iter(items)
    for first in items:
        chunk = [first, *islice(items, chunk_size - 1)]
        if len(chunk) >= chunk_size or include_remainder:
            yield chunk


def iter_rechunk(chunks, chunk_size: int, include_remainder=True):
    """
    takes in chunks and returns chunks of a new size.
    - Does not evaluate unneeded chunks
    """
    return iter_chunks(
        (item for chunk in chunks for item in chunk),  # flatten chunks
        chunk_size=chunk_size,
        include_remainder=include_remainder,
    )


def map_all(fn, *arg_lists, starmap: bool = True, collect_returned: bool = False, common_kwargs: dict = None):
    # TODO: not actually an iterator
    assert arg_lists, "an empty list of args was passed"
    # check all lengths are the same
    num = len(arg_lists[0])
    assert num > 0
    assert all(len(items) == num for items in arg_lists)
    # update kwargs
    if common_kwargs is None:
        common_kwargs = {}
    # map everything
    if starmap:
        results = (fn(*args, **common_kwargs) for args in zip(*arg_lists))
    else:
        results = (fn(args, **common_kwargs) for args in zip(*arg_lists))
    # zip everything
    if collect_returned:
        return tuple(zip(*results))
    else:
        return tuple(results)


def collect_dicts(results: List[dict]):
    # collect everything
    keys = results[0].keys()
    values = zip(*([result[k] for k in keys] for result in results))
    return {k: list(v) for k, v in zip(keys, values)}


def aggregate_dict(results: dict, reduction="mean"):
    # TODO: this shouldn't be here
    assert reduction == "mean", "mean is the only mode supported"
    return {k: sum(v) / len(v) for k, v in results.items()}


# ========================================================================= #
# Base Class                                                                #
# ========================================================================= #


class LengthIter(Sequence):
    def __iter__(self):
        # this takes priority over __getitem__, otherwise __getitem__ would need to
        # raise an IndexError if out of bounds to signal the end of iteration
        yield from (self[i] for i in range(len(self)))

    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, item):
        raise NotImplementedError()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import torch

# ========================================================================= #
# pytorch math correlation functions                                        #
# ========================================================================= #


def torch_cov_matrix(xs: torch.Tensor):
    """
    Calculate the covariance matrix of multiple samples (N) of random vectors of size (X)
    https://en.wikipedia.org/wiki/Covariance_matrix
    - The input shape is: (N, X)
    - The output shape is: (X, X)

    This should be the same as:
        np.cov(xs, rowvar=False, ddof=0)
    """
    # NOTE:
    #   torch.mm is strict matrix multiplication
    #   however if we multiply arrays with broadcasting:
    #   size(3, 1) * size(1, 2) -> size(3, 2)  # broadcast, not matmul
    #   size(1, 3) * size(2, 1) -> size(2, 3)  # broadcast, not matmul
    # CHECK:
    assert xs.ndim == 2  # (N, X)
    Rxx = torch.mean(xs[:, :, None] * xs[:, None, :], dim=0)  # (X, X)
    ux = torch.mean(xs, dim=0)  # (X,)
    Kxx = Rxx - (ux[:, None] * ux[None, :])  # (X, X)
    return Kxx


def torch_corr_matrix(xs: torch.Tensor):
    """
    Calculate the pearson's correlation matrix of multiple samples (N) of random vectors of size (X)
    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
    https://en.wikipedia.org/wiki/Covariance_matrix
    - The input shape is: (N, X)
    - The output shape is: (X, X)

    This should be the same as:
        np.corrcoef(xs, rowvar=False, ddof=0)
    """
    Kxx = torch_cov_matrix(xs)
    diag_Kxx = torch.rsqrt(torch.diagonal(Kxx))
    corr = Kxx * (diag_Kxx[:, None] * diag_Kxx[None, :])
    return corr


def torch_rank_corr_matrix(xs: torch.Tensor):
    """
    Calculate the spearman's rank correlation matrix of multiple samples (N) of random vectors of size (X)
    https://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient
    - The input shape is: (N, X)
    - The output shape is: (X, X)

    Pearson's correlation measures linear relationships
    Spearman's correlation measures monotonic relationships (whether linear or not)
    - defined in terms of the pearson's correlation matrix of the rank variables

    TODO: check, be careful of repeated values, this might not give the correct input?
    """
    rs = torch.argsort(xs, dim=0, descending=False)
    return torch_corr_matrix(rs.to(xs.dtype))


# aliases
torch_pearsons_corr_matrix = torch_corr_matrix
torch_spearmans_corr_matrix = torch_rank_corr_matrix


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from typing import Dict
from typing import NoReturn
from typing import Optional
from typing import Union

import numpy as np

from disent.dataset.data import NumpyFileGroundTruthData
from disent.dataset.util.datafile import DataFileHashed
from disent.util.inout.paths import modify_name_keep_ext

log = logging.getLogger(__name__)


# ========================================================================= #
# teapots 3d data processing                                                #
# ========================================================================= #


def resave_teapots3d_as_uint8(orig_file: str, new_file: str, overwrite: bool = False):
    # load data into memory ~10GB
    # -- by default this array is stored as uint32, instead of uint8
    log.debug("loading teapots data into memory, this may take a while...")
    imgs = np.load(orig_file)["images"]
    log.debug("loaded teapots data into memory!")
    # checks
    log.debug("checking teapots data...")
    assert imgs.dtype == "int32"
    assert imgs.shape == (200_000, 64, 64, 3)
    assert imgs.max() == 255
    assert imgs.min() == 0
    log.debug("checked teapots data!")
    # convert the values
    log.debug("converting teapots data to uint8...")
    imgs = imgs.astype("uint8")
    log.debug("converted teapots data!")
    # save the array
    from disent.dataset.util.formats.npz import save_dataset_array

    log.debug("saving convert teapots data...")
    save_dataset_array(imgs, new_file, overwrite=overwrite, save_key="images")
    log.debug("saved convert teapots data!")


# ========================================================================= #
# teapots 3d data files                                                     #
# ========================================================================= #


class DataFileTeapots3dInt32(DataFileHashed):
    # TODO: add a version of this file that automatically unpacks the original zip file?

    def _prepare(self, out_dir: str, out_file: str) -> NoReturn:
        if not os.path.exists(out_file):
            raise FileNotFoundError(
                f"Please download the Teapots3D dataset to: {repr(out_file)}"
                f'\nThe original repository is: {repr("https://github.com/cianeastwood/qedr")}'
                f'\nThe original download link is: {repr("https://www.dropbox.com/s/woeyomxuylqu7tx/edinburgh_teapots.zip?dl=0")}'
            )


class DataFileTeapots3dUint8(DataFileHashed):
    def __init__(
        self,
        teapots3d_datafile: DataFileTeapots3dInt32,
        # - convert file name
        out_hash: Optional[Union[str, Dict[str, str]]],
        out_name: Optional[str] = None,
        # - hash settings
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self._teapots3dfile = teapots3d_datafile
        super().__init__(
            file_name=modify_name_keep_ext(self._teapots3dfile.out_name, suffix=f"_uint8")
            if (out_name is None)
            else out_name,
            file_hash=out_hash,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )

    def _prepare(self, out_dir: str, out_file: str):
        log.debug("Preparing Orig Teapots3d Data:")
        cars3d_path = self._teapots3dfile.prepare(out_dir)
        log.debug("Converting Teapots3d Data to Uint8:")
        resave_teapots3d_as_uint8(orig_file=cars3d_path, new_file=out_file, overwrite=True)


# ========================================================================= #
# teapots 3d dataset                                                        #
# ========================================================================= #


class Teapots3dData(NumpyFileGroundTruthData):
    """
    Teapots3D Dataset
    -  A Framework for the Quantitative Evaluation of Disentangled Representations
       * https://openreview.net/forum?id=By-7dz-AZ
       * https://github.com/cianeastwood/qedr

    Manual Download Link:
    - https://www.dropbox.com/s/woeyomxuylqu7tx/edinburgh_teapots.zip?dl=0

    NOTE:
    - This dataset is generated from ground-truth factors, HOWEVER, each datapoint
      is randomly sampled. This dataset is NOT a typical grid-search over ground-truth factors
      which means that we cannot create a StateSpace object over this dataset.
    """

    name = "edinburgh_teapots"

    factor_names = ("azimuth", "elevation", "red", "green", "blue")
    factor_sizes = (
        ...,
        ...,
        ...,
        ...,
        ...,
    )  # TOTAL: 200_000 -- TODO: this is invalid, we cannot actually generate a StateSpace object over this dataset!
    img_shape = (64, 64, 3)

    datafile = DataFileTeapots3dUint8(
        teapots3d_datafile=DataFileTeapots3dInt32(
            file_name="teapots.npz",
            file_hash={"full": "9b58d66a382d01f4477e33520f1fa503", "fast": "12c889e001c205d0bafa59dfff114102"},
        ),
        out_hash={"full": "e64207ee443030d310500d762f0d1dfd", "fast": "7fbca0223c27e055d35b6d5af720f108"},
        out_name="teapots_uint8.npz",
    )

    # override
    data_key = "images"


# ========================================================================= #
# main                                                                      #
# ========================================================================= #


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    data = Teapots3dData(data_root="~/Downloads", prepare=True)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Callable
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Union

import hydra
from omegaconf import DictConfig
from omegaconf import OmegaConf

from experiment.util.path_utils import get_current_experiment_number
from experiment.util.path_utils import make_current_experiment_dir
from experiment.util.run_utils import log_error_and_exit

log = logging.getLogger(__name__)


# ========================================================================= #
# VARS                                                                      #
# ========================================================================= #


# experiment/util/_hydra_searchpath_plugin_
PLUGIN_NAMESPACE = os.path.abspath(os.path.join(__file__, "..", "_hydra_searchpath_plugin_"))

# experiment/config
EXP_CONFIG_DIR = os.path.abspath(os.path.join(__file__, "../..", "config"))

# list of configs
_DISENT_CONFIG_DIRS: List[str] = None


# ========================================================================= #
# PATCHING                                                                  #
# ========================================================================= #


def register_searchpath_plugin(
    search_dir_main: str = EXP_CONFIG_DIR,
    search_dirs_prepend: Optional[Union[str, List[str]]] = None,
    search_dirs_append: Optional[Union[str, List[str]]] = None,
):
    """
     Patch Hydra:
     1. sets the default search path to `experiment/config`
     2. add to the search path with the `DISENT_CONFIGS_PREPEND` and `DISENT_CONFIGS_APPEND` environment variables
    NOTE: --config-dir has lower priority than all these, --config-path has higher priority.

     This function can safely be called multiple times
         - unless other functions modify these same libs which is unlikely!
    """
    # normalise the config paths
    if search_dirs_prepend is None:
        search_dirs_prepend = []
    if search_dirs_append is None:
        search_dirs_append = []
    if isinstance(search_dirs_prepend, str):
        search_dirs_prepend = [search_dirs_prepend]
    if isinstance(search_dirs_append, str):
        search_dirs_append = [search_dirs_append]
    assert isinstance(search_dirs_prepend, (tuple, list)) and all(
        (isinstance(d, str) and d) for d in search_dirs_prepend
    ), f"`search_dirs_prepend` must be a list or tuple of non-empty path strings to directories, got: {repr(search_dirs_prepend)}"
    assert isinstance(search_dirs_append, (tuple, list)) and all(
        (isinstance(d, str) and d) for d in search_dirs_append
    ), f"`search_dirs_append` must be a list or tuple of non-empty path strings to directories, got: {repr(search_dirs_append)}"
    assert (
        isinstance(search_dir_main, str) and search_dir_main
    ), f"`search_dir_main` must be a non-empty path string to a directory, got: {repr(search_dir_main)}"
    # get dirs
    config_dirs = [*search_dirs_prepend, search_dir_main, *search_dirs_append]

    # check that it is the same as what has previously been registered, otherwise set the directories!
    global _DISENT_CONFIG_DIRS
    if _DISENT_CONFIG_DIRS is None:
        _DISENT_CONFIG_DIRS = config_dirs
    else:
        assert (
            _DISENT_CONFIG_DIRS == config_dirs
        ), f"Config dirs have already been registered, on additional calls, registered dirs must be the same as previously values!\n- existing: {_DISENT_CONFIG_DIRS}\n- registered: {config_dirs}"

    # register the experiment's search path plugin with disent, using hydras auto-detection
    # of folders named `hydra_plugins` contained insided `namespace packages` or rather
    # packages that are in the `PYTHONPATH` or `sys.path`
    #   1. sets the default search path to those registered above
    #   2. add to the search path with the `DISENT_CONFIGS_PREPEND` and `DISENT_CONFIGS_APPEND` environment variables
    #      NOTE: --config-dir has lower priority than all these, --config-path has higher priority.
    if PLUGIN_NAMESPACE not in sys.path:
        sys.path.insert(0, PLUGIN_NAMESPACE)


def register_hydra_resolvers():
    """
    Patch OmegaConf, enabling various config resolvers:
        - enable the ${exit:<msg>} resolver for omegaconf/hydra
        - enable the ${exp_num:<root_dir>} and ${exp_dir:<root_dir>,<name>} resolvers to detect the experiment number
        - enable the ${fmt:<strfmt>,<kwargs...>} resolver that wraps `str.format`
        - enable the ${abspath:<path>} resolver that wraps `hydra.utils.to_absolute_path` formatting relative paths in relation to the original working directory
        - enable the ${rsync_dir:<src>/<name>,<dst>/<name>} resolver that returns `<dst>/<name>`, but first rsync's the two directories!

    This function can safely be called multiple times
        - unless other functions modify these same libs which is unlikely!
    """

    # register a custom OmegaConf resolver that allows us to put in a ${exit:msg} that exits the program
    # - if we don't register this, the program will still fail because we have an unknown
    #   resolver. This just prettifies the output.
    if not OmegaConf.has_resolver("exit"):

        class ConfigurationError(Exception):
            pass

        # resolver function
        def _error_resolver(msg: str):
            raise ConfigurationError(msg)

        # patch omegaconf for hydra
        OmegaConf.register_new_resolver("exit", _error_resolver)

    # register a custom OmegaConf resolver that allows us to get the next experiment number from a directory
    # - ${run_num:<root_dir>} returns the current experiment number
    if not OmegaConf.has_resolver("exp_num"):
        OmegaConf.register_new_resolver("exp_num", get_current_experiment_number)
    # - ${run_dir:<root_dir>,<name>} returns the current experiment folder with the name appended
    if not OmegaConf.has_resolver("exp_dir"):
        OmegaConf.register_new_resolver("exp_dir", make_current_experiment_dir)

    # register a function that pads an integer to a specified length
    # - ${fmt:"{:04d}",42} -> "0042"
    if not OmegaConf.has_resolver("fmt"):
        OmegaConf.register_new_resolver("fmt", str.format)

    # register hydra helper functions
    # - ${abspath:<rel_path>} convert a relative path to an abs path using the original hydra working directory, not the changed experiment dir.
    if not OmegaConf.has_resolver("abspath"):
        OmegaConf.register_new_resolver("abspath", hydra.utils.to_absolute_path)

    # registry copy directory function
    # - useful if datasets are already prepared on a shared drive and need to be copied to a temp drive for example!
    if not OmegaConf.has_resolver("rsync_dir"):

        def rsync_dir(src: str, dst: str) -> str:
            src, dst = Path(src), Path(dst)
            # checks
            assert src.name and src.is_absolute(), f"src path must be absolute and not the root: {repr(str(src))}"
            assert dst.name and dst.is_absolute(), f"dst path must be absolute and not the root: {repr(str(dst))}"
            assert (
                src.name == dst.name
            ), f"src and dst paths must point to dirs with the same names: src.name={repr(src.name)}, dst.name={repr(dst.name)}"
            # synchronize dirs
            logging.info(f"rsync files:\n- src={repr(str(src))}\n- dst={repr(str(dst))}")
            # create the parent dir and copy files into the parent
            dst.parent.mkdir(parents=True, exist_ok=True)
            returncode = subprocess.Popen(["rsync", "-avh", str(src), str(dst.parent)]).wait()
            if returncode != 0:
                raise RuntimeError("Failed to rsync files!")
            # return the destination dir
            return str(dst)

        # REGISTER
        OmegaConf.register_new_resolver("rsync_dir", rsync_dir)


# ========================================================================= #
# RUN HYDRA                                                                 #
# ========================================================================= #


def patch_hydra(
    # config search path
    search_dir_main: str = EXP_CONFIG_DIR,
    search_dirs_prepend: Optional[Union[str, List[str]]] = None,
    search_dirs_append: Optional[Union[str, List[str]]] = None,
):
    # Patch Hydra and OmegaConf:
    register_searchpath_plugin(
        search_dir_main=search_dir_main, search_dirs_prepend=search_dirs_prepend, search_dirs_append=search_dirs_append
    )
    register_hydra_resolvers()


def hydra_main(
    callback: Callable[[DictConfig], NoReturn],
    config_name: str = "config",
    # config search path
    search_dir_main: str = EXP_CONFIG_DIR,
    search_dirs_prepend: Optional[Union[str, List[str]]] = None,
    search_dirs_append: Optional[Union[str, List[str]]] = None,
    # logging
    log_level: Optional[int] = logging.INFO,
    log_exc_info_callback: bool = True,
    log_exc_info_hydra: bool = False,
):
    # manually set log level before hydra initialises!
    if log_level is not None:
        logging.basicConfig(level=log_level)

    # Patch Hydra and OmegaConf:
    patch_hydra(
        search_dir_main=search_dir_main, search_dirs_prepend=search_dirs_prepend, search_dirs_append=search_dirs_append
    )

    @hydra.main(config_path=None, config_name=config_name)
    def _hydra_main(cfg: DictConfig):
        try:
            callback(cfg)
        except Exception as e:
            log_error_and_exit(err_type="experiment error", err_msg=str(e), exc_info=log_exc_info_callback)
        except:
            log_error_and_exit(err_type="experiment error", err_msg="<UNKNOWN>", exc_info=log_exc_info_callback)

    try:
        _hydra_main()
    except KeyboardInterrupt as e:
        log_error_and_exit(err_type="interrupted", err_msg=str(e), exc_info=False)
    except Exception as e:
        log_error_and_exit(err_type="hydra error", err_msg=str(e), exc_info=log_exc_info_hydra)
    except:
        log_error_and_exit(err_type="hydra error", err_msg="<UNKNOWN>", exc_info=log_exc_info_hydra)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple

import torch

from disent.frameworks.ae._supervised__adaneg_tae import AdaNegTripletAe
from disent.frameworks.vae._supervised__adaneg_tvae import AdaNegTripletVae
from disent.frameworks.vae._unsupervised__dotvae import DataOverlapMixin

log = logging.getLogger(__name__)


# ========================================================================= #
# Data Overlap Triplet AE                                                  #
# ========================================================================= #


class DataOverlapTripletAe(AdaNegTripletAe, DataOverlapMixin):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    This is the unsupervised version of the Adaptive Triplet Loss for Auto-Encoders not VAEs
    - Triplets are sampled from training minibatches and are ordered using some distance
      function taken directly over the datapoints.
    - Adaptive Triplet Loss is used to guide distance learning and encourage disentanglement
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(AdaNegTripletAe.cfg, DataOverlapMixin.cfg):
        pass

    def __init__(self, model: "AutoEncoder", cfg: cfg = None, batch_augment=None):
        super().__init__(model=model, cfg=cfg, batch_augment=batch_augment)
        # initialise mixin
        self.init_data_overlap_mixin()

    def hook_ae_compute_ave_aug_loss(
        self, zs: Sequence[torch.Tensor], xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[torch.Tensor, Dict[str, Any]]:
        [z], [x_targ_orig] = zs, xs_targ
        # 1. randomly generate and mine triplets using augmented versions of the inputs
        a_idxs, p_idxs, n_idxs = self.random_mined_triplets(x_targ_orig=x_targ_orig)
        # 2. compute triplet loss
        loss, loss_log = AdaNegTripletVae.estimate_ada_triplet_loss_from_zs(
            zs=[z[idxs] for idxs in (a_idxs, p_idxs, n_idxs)],
            cfg=self.cfg,
        )
        return loss, {
            **loss_log,
        }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import torch

# ========================================================================= #
# Discrete Cosine Transform                                                 #
# ========================================================================= #


def _flatten_dim_to_end(input, dim):
    # get shape
    s = input.shape
    n = s[dim]
    # do operation
    x = torch.moveaxis(input, dim, -1)
    x = x.reshape(-1, n)
    return x, s, n


def _unflatten_dim_to_end(input, dim, shape):
    # get intermediate shape
    s = list(shape)
    s.append(s.pop(dim))
    # undo operation
    x = input.reshape(*s)
    x = torch.moveaxis(x, -1, dim)
    return x


def torch_dct(x, dim=-1):
    """
    Discrete Cosine Transform (DCT) Type II
    """
    x, x_shape, n = _flatten_dim_to_end(x, dim=dim)
    if n % 2 != 0:
        raise ValueError(
            f"dct does not support odd sized dimension! trying to compute dct over dimension: {dim} of tensor with shape: {x_shape}"
        )

    # concatenate even and odd offsets
    v_evn = x[:, 0::2]
    v_odd = x[:, 1::2].flip([1])
    v = torch.cat([v_evn, v_odd], dim=-1)

    # fast fourier transform
    fft = torch.fft.fft(v)

    # compute real & imaginary forward weights
    k = torch.arange(n, dtype=x.dtype, device=x.device) * (-np.pi / (2 * n))
    k = k[None, :]
    wr = torch.cos(k) * 2
    wi = torch.sin(k) * 2

    # compute dct
    dct = torch.real(fft) * wr - torch.imag(fft) * wi

    # restore shape
    return _unflatten_dim_to_end(dct, dim, x_shape)


def torch_idct(dct, dim=-1):
    """
    Inverse Discrete Cosine Transform (Inverse DCT) Type III
    """
    dct, dct_shape, n = _flatten_dim_to_end(dct, dim=dim)
    if n % 2 != 0:
        raise ValueError(
            f"idct does not support odd sized dimension! trying to compute idct over dimension: {dim} of tensor with shape: {dct_shape}"
        )

    # compute real & imaginary backward weights
    k = torch.arange(n, dtype=dct.dtype, device=dct.device) * (np.pi / (2 * n))
    k = k[None, :]
    wr = torch.cos(k) / 2
    wi = torch.sin(k) / 2

    dct_real = dct
    dct_imag = torch.cat([0 * dct_real[:, :1], -dct_real[:, 1:].flip([1])], dim=-1)

    fft_r = dct_real * wr - dct_imag * wi
    fft_i = dct_real * wi + dct_imag * wr
    # to complex number
    fft = torch.view_as_complex(torch.stack([fft_r, fft_i], dim=-1))

    # inverse fast fourier transform
    v = torch.fft.ifft(fft)
    v = torch.real(v)

    # undo even and odd offsets
    x = torch.zeros_like(dct)
    x[:, 0::2] = v[:, : (n + 1) // 2]  # (N+1)//2 == N-(N//2)
    x[:, 1::2] += v[:, (n + 0) // 2 :].flip([1])

    # restore shape
    return _unflatten_dim_to_end(x, dim, dct_shape)


def torch_dct2(x, dim1=-1, dim2=-2):
    d = torch_dct(x, dim=dim1)
    d = torch_dct(d, dim=dim2)
    return d


def torch_idct2(d, dim1=-1, dim2=-2):
    x = torch_idct(d, dim=dim2)
    x = torch_idct(x, dim=dim1)
    return x


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Tuple
from typing import final

import torch

from disent.frameworks import DisentFramework
from disent.frameworks.helper.reconstructions import ReconLossHandler
from disent.frameworks.helper.reconstructions import make_reconstruction_loss
from disent.model import AutoEncoder

log = logging.getLogger(__name__)


# ========================================================================= #
# framework_vae                                                             #
# ========================================================================= #


class _AeAndVaeMixin(DisentFramework):
    """
    Base class containing common logic for both the Ae and Vae classes.
    -- This is private because handling of both classes needs to be conducted differently.
       An instance of a Vae should not be an instance of an Ae and vice-versa
    """

    @dataclass
    class cfg(DisentFramework.cfg):
        recon_loss: str = "mse"
        # multiple reduction modes exist for the various loss components.
        # - 'sum': sum over the entire batch
        # - 'mean': mean over the entire batch
        # - 'mean_sum': sum each observation, returning the mean sum over the batch
        loss_reduction: str = "mean"
        # disable various components
        detach_decoder: bool = False
        disable_rec_loss: bool = False
        disable_aug_loss: bool = False

    # --------------------------------------------------------------------- #
    # AE/VAE Attributes                                                     #
    # --------------------------------------------------------------------- #

    @property
    def REQUIRED_Z_MULTIPLIER(self) -> int:
        raise NotImplementedError

    @property
    def REQUIRED_OBS(self) -> int:
        raise NotImplementedError

    @final
    @property
    def recon_handler(self) -> ReconLossHandler:
        return self.__recon_handler

    # --------------------------------------------------------------------- #
    # AE/VAE Init                                                           #
    # --------------------------------------------------------------------- #

    # attributes provided by this class and initialised in _init_ae_mixin
    _model: AutoEncoder
    __recon_handler: ReconLossHandler

    def _init_ae_mixin(self, model: AutoEncoder):
        # vae model
        self._model = model
        # check the model
        assert isinstance(
            self._model, AutoEncoder
        ), f"model must be an instance of {AutoEncoder.__name__}, got: {type(model)}"
        assert (
            self._model.z_multiplier == self.REQUIRED_Z_MULTIPLIER
        ), f"model z_multiplier is {repr(self._model.z_multiplier)} but {self.__class__.__name__} requires that it is: {repr(self.REQUIRED_Z_MULTIPLIER)}"
        # recon loss & activation fn
        self.__recon_handler: ReconLossHandler = make_reconstruction_loss(
            self.cfg.recon_loss, reduction=self.cfg.loss_reduction
        )

    # --------------------------------------------------------------------- #
    # AE/VAE Training Step Helper                                           #
    # --------------------------------------------------------------------- #

    @final
    def _get_xs_and_targs(
        self, batch: Dict[str, Tuple[torch.Tensor, ...]], batch_idx
    ) -> Tuple[Tuple[torch.Tensor, ...], Tuple[torch.Tensor, ...]]:
        xs_targ = batch["x_targ"]
        if "x" not in batch:
            # TODO: re-enable this warning but only ever print once!
            # warnings.warn('dataset does not have input: x -> x_targ using target as input: x_targ -> x_targ')
            xs = xs_targ
        else:
            xs = batch["x"]
        # check that we have the correct number of inputs
        if (len(xs) != self.REQUIRED_OBS) or (len(xs_targ) != self.REQUIRED_OBS):
            log.warning(
                f"batch len(xs)={len(xs)} and len(xs_targ)={len(xs_targ)} observation count mismatch, requires: {self.REQUIRED_OBS}"
            )
        # done
        return xs, xs_targ

    # --------------------------------------------------------------------- #
    # AE/VAE Model Utility Functions (Visualisation)                        #
    # --------------------------------------------------------------------- #

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """Get the deterministic latent representation (useful for visualisation)"""
        raise NotImplementedError

    def decode(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into reconstruction x_recon (useful for visualisation)"""
        raise NotImplementedError

    def forward(self, batch: torch.Tensor) -> torch.Tensor:
        """Feed through the full deterministic model (useful for visualisation)"""
        raise NotImplementedError

    # --------------------------------------------------------------------- #
    # AE/VAE Model Utility Functions (Training)                             #
    # --------------------------------------------------------------------- #

    def decode_partial(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into partial reconstructions that exclude the final activation if there is one."""
        raise NotImplementedError

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        if not checkpoint.get(
            "hyper_parameters"
        ):  # if the logger did not register hyperparameters they're set here manually
            checkpoint["hyper_parameters"] = {}
        checkpoint["hyper_parameters"]["model"] = self._model
        checkpoint["hyper_parameters"]["cfg"] = self.cfg
        checkpoint["hyper_parameters"]["batch_augment"] = self._batch_augment
        return super().on_save_checkpoint(checkpoint)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Optional

import torch
from torch import nn

from disent.util.strings import colors as c

log = logging.getLogger(__name__)


# ========================================================================= #
# Basic Weight Initialisation                                               #
# ========================================================================= #


_WEIGHT_INIT_FNS = {
    "xavier_uniform": lambda weight: nn.init.xavier_uniform_(weight, gain=1.0),  # gain=1
    "xavier_normal": lambda weight: nn.init.xavier_normal_(weight, gain=1.0),  # gain=1
    "xavier_normal__0.1": lambda weight: nn.init.xavier_normal_(weight, gain=0.1),  # gain=0.1
    # kaiming -- also known as "He initialisation"
    "kaiming_uniform": lambda weight: nn.init.kaiming_uniform_(
        weight, a=0, mode="fan_in", nonlinearity="relu"
    ),  # fan_in, relu
    "kaiming_normal": lambda weight: nn.init.kaiming_normal_(
        weight, a=0, mode="fan_in", nonlinearity="relu"
    ),  # fan_in, relu
    "kaiming_normal__fan_out": lambda weight: nn.init.kaiming_normal_(
        weight, a=0, mode="fan_out", nonlinearity="relu"
    ),  # fan_in, relu
    # other
    "orthogonal": lambda weight: nn.init.orthogonal_(weight, gain=1),  # gain=1
    "normal": lambda weight: nn.init.normal_(weight, mean=0.0, std=1.0),  # gain=1
    "normal__0.1": lambda weight: nn.init.normal_(weight, mean=0.0, std=0.1),  # gain=0.1
    "normal__0.01": lambda weight: nn.init.normal_(weight, mean=0.0, std=0.01),  # gain=0.01
    "normal__0.001": lambda weight: nn.init.normal_(weight, mean=0.0, std=0.001),  # gain=0.01
}


# TODO: clean this up! this is terrible...
def init_model_weights(model: nn.Module, mode: Optional[str] = "xavier_normal", log_level=logging.INFO) -> nn.Module:
    count = 0

    # get default mode
    if mode is None:
        mode = "default"

    def _apply_init_weights(m):
        nonlocal count
        init, count = False, count + 1

        # actually initialise!
        if mode == "default":
            pass
        elif mode in _WEIGHT_INIT_FNS:
            if isinstance(m, (nn.Linear, nn.Conv2d, nn.ConvTranspose2d)):
                _WEIGHT_INIT_FNS[mode](m.weight)
                nn.init.zeros_(m.bias)
                init = True
        else:
            raise KeyError(
                f'Unknown init mode: {repr(mode)}, valid modes are: {["default"] + sorted(_WEIGHT_INIT_FNS)}'
            )

        # print messages
        if init:
            log.log(log_level, f"| {count:03d} {c.lGRN}INIT{c.RST}: {m.__class__.__name__}")
        else:
            log.log(log_level, f"| {count:03d} {c.lRED}SKIP{c.RST}: {m.__class__.__name__}")

    log.log(log_level, f"Initialising Model Layers: {mode}")
    model.apply(_apply_init_weights)

    return model


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


# ========================================================================= #
# Working With Arbitrary Precision Integers                                 #
# ========================================================================= #


def gcd(a: int, b: int) -> int:
    """
    Compute the greatest common divisor of a and b
    TODO: not actually sure if this returns the correct values for zero or negative inputs?
    """
    assert isinstance(a, int), f"number must be an int, got: {type(a)}"
    assert isinstance(b, int), f"number must be an int, got: {type(b)}"
    while b > 0:
        a, b = b, a % b
    return a


def lcm(a: int, b: int) -> int:
    """
    Compute the lowest common multiple of a and b
    TODO: not actually sure if this returns the correct values for zero or negative inputs?
    """
    return (a * b) // gcd(a, b)


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from numbers import Number
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import final

import torch
from torch.distributions import Distribution

from disent.frameworks._ae_mixin import _AeAndVaeMixin
from disent.frameworks.helper.latent_distributions import LatentDistsHandler
from disent.frameworks.helper.latent_distributions import make_latent_distribution
from disent.frameworks.helper.util import detach_all
from disent.util.iters import map_all

# ========================================================================= #
# framework_vae                                                             #
# ========================================================================= #


class Vae(_AeAndVaeMixin):
    """
    Variational Auto Encoder
    https://arxiv.org/abs/1312.6114
    ------------------------

    This VAE implementation supports multiple inputs in parallel. Each input
    is fed through the VAE on its own, and the loss from each step is averaged
    together at the end. This is effectively the same as increasing the batch
    size and scaling down the loss. The reason this VAE implementation supports
    the variable number of input args is to allow one common framework to be
    the parent of all child VAEs.

    Child classes can implement various hooks to override or add additional
    functionality to the VAE. Most common VAE derivatives can be implemented
    by simply adding functionality using these hooks, and changing the required
    number of input arguments using `REQUIRED_OBS`.

    - HOOKS:
        * `hook_intercept_ds`
        * `hook_compute_ave_aug_loss` (NB: not the same as `hook_ae_compute_ave_aug_loss` from AEs)

    - OVERRIDES:
        * `compute_ave_recon_loss`
        * `compute_ave_reg_loss`

    For example:
    -> implementing `hook_compute_ave_aug_loss` and setting `REQUIRED_OBS=3`
       we can easily implement a Triplet VAE.
    -> implementing `hook_intercept_ds` and then setting `REQUIRED_OBS=2`
       we can easily implement the Adaptive VAE style frameworks.

    TODO: allow hooks to be registered? simply build up new frameworks?
          Vae(recon_loss=DfcLoss())
          Vae(hook_intercept_ds=AdaptiveAveraging(mode='gvae'), required_obs=2)
          Vae(hook_compute_ave_aug_loss=TripletLoss(), required_obs=3)
    """

    # overrides
    REQUIRED_Z_MULTIPLIER = 2
    REQUIRED_OBS = 1

    @dataclass
    class cfg(_AeAndVaeMixin.cfg):
        # latent distribution settings
        latent_distribution: str = "normal"
        kl_loss_mode: str = "direct"
        # disable various components
        disable_reg_loss: bool = False

    def __init__(self, model: "AutoEncoder", cfg: cfg = None, batch_augment=None):
        # required_z_multiplier
        super().__init__(cfg=cfg, batch_augment=batch_augment)
        # initialise the auto-encoder mixin (recon handler, model, enc, dec, etc.)
        self._init_ae_mixin(model=model)
        # vae distribution
        self.__latents_handler = make_latent_distribution(
            self.cfg.latent_distribution, kl_mode=self.cfg.kl_loss_mode, reduction=self.cfg.loss_reduction
        )

    @final
    @property
    def latents_handler(self) -> LatentDistsHandler:
        return self.__latents_handler

    # --------------------------------------------------------------------- #
    # VAE Training Step                                                     #
    # --------------------------------------------------------------------- #

    @final
    def do_training_step(self, batch, batch_idx):
        xs, xs_targ = self._get_xs_and_targs(batch, batch_idx)

        # FORWARD
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        # latent distribution parameterizations
        ds_posterior, ds_prior = map_all(self.encode_dists, xs, collect_returned=True)
        # [HOOK] intercept latent parameterizations
        ds_posterior, ds_prior, logs_intercept_ds = self.hook_intercept_ds(ds_posterior, ds_prior)
        # sample from dists
        zs_sampled = tuple(d.rsample() for d in ds_posterior)
        # reconstruct without the final activation
        xs_partial_recon = map_all(self.decode_partial, detach_all(zs_sampled, if_=self.cfg.detach_decoder))
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

        # LOSS
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        # compute all the recon losses
        recon_loss, logs_recon = self.compute_ave_recon_loss(xs_partial_recon, xs_targ)
        # compute all the regularization losses
        reg_loss, logs_reg = self.compute_ave_reg_loss(ds_posterior, ds_prior, zs_sampled)
        # [HOOK] augment loss
        aug_loss, logs_aug = self.hook_compute_ave_aug_loss(
            ds_posterior=ds_posterior,
            ds_prior=ds_prior,
            zs_sampled=zs_sampled,
            xs_partial_recon=xs_partial_recon,
            xs_targ=xs_targ,
        )
        # compute combined loss
        loss = 0
        if not self.cfg.disable_rec_loss:
            loss += recon_loss
        if not self.cfg.disable_aug_loss:
            loss += aug_loss
        if not self.cfg.disable_reg_loss:
            loss += reg_loss
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

        # log general
        self.log_dict(
            {
                **logs_intercept_ds,
                **logs_recon,
                **logs_reg,
                **logs_aug,
            }
        )

        # log progress bar
        self.log_dict(
            {
                "recon_loss": float(recon_loss),
                "reg_loss": float(reg_loss),
                "aug_loss": float(aug_loss),
            },
            prog_bar=True,
        )

        # return values
        return loss

    # --------------------------------------------------------------------- #
    # Overrideable Hooks                                                    #
    # --------------------------------------------------------------------- #

    def hook_intercept_ds(
        self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution]
    ) -> Tuple[Sequence[Distribution], Sequence[Distribution], Dict[str, Any]]:
        return ds_posterior, ds_prior, {}

    def hook_compute_ave_aug_loss(
        self,
        ds_posterior: Sequence[Distribution],
        ds_prior: Sequence[Distribution],
        zs_sampled: Sequence[torch.Tensor],
        xs_partial_recon: Sequence[torch.Tensor],
        xs_targ: Sequence[torch.Tensor],
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        return 0, {}

    def compute_ave_recon_loss(
        self, xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        # compute reconstruction loss
        pixel_loss = self.recon_handler.compute_ave_loss_from_partial(xs_partial_recon, xs_targ)
        # return logs
        return pixel_loss, {"pixel_loss": pixel_loss}

    def compute_ave_reg_loss(
        self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution], zs_sampled: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        # compute regularization loss (kl divergence)
        kl_loss = self.latents_handler.compute_ave_kl_loss(ds_posterior, ds_prior, zs_sampled)
        # return logs
        return kl_loss, {
            "kl_loss": kl_loss,
        }

    # --------------------------------------------------------------------- #
    # VAE Model Utility Functions (Visualisation)                           #
    # --------------------------------------------------------------------- #

    @final
    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """Get the deterministic latent representation (useful for visualisation)"""
        z_raw = self._model.encode(x, chunk=True)
        z = self.latents_handler.encoding_to_representation(z_raw)
        return z

    @final
    def decode(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into reconstruction x_recon (useful for visualisation)"""
        return self.recon_handler.activate(self._model.decode(z))

    @final
    def forward(self, batch: torch.Tensor) -> torch.Tensor:
        """Feed through the full deterministic model (useful for visualisation)"""
        return self.decode(self.encode(batch))

    # --------------------------------------------------------------------- #
    # VAE Model Utility Functions (Training)                                #
    # --------------------------------------------------------------------- #

    @final
    def encode_dists(self, x: torch.Tensor) -> Tuple[Distribution, Distribution]:
        """Get parametrisations of the latent distributions, which are sampled from during training."""
        z_raw = self._model.encode(x, chunk=True)
        z_posterior, z_prior = self.latents_handler.encoding_to_dists(z_raw)
        return z_posterior, z_prior

    @final
    def decode_partial(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into partial reconstructions that exclude the final activation if there is one."""
        return self._model.decode(z)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from disent.dataset.data import BaseEpisodesData
from disent.dataset.sampling._base import BaseDisentSampler
from disent.util.math.random import sample_radius as sample_radius_fn

# ========================================================================= #
# Episode Sampler                                                           #
# ========================================================================= #


class RandomEpisodeSampler(BaseDisentSampler):
    def uninit_copy(self) -> "RandomEpisodeSampler":
        return RandomEpisodeSampler(
            num_samples=self.num_samples,
            sample_radius=self._sample_radius,
        )

    def __init__(self, num_samples=1, sample_radius=None):
        super().__init__(num_samples=num_samples)
        self._sample_radius = sample_radius
        # checks
        if sample_radius is not None:
            if sample_radius >= 0:
                if num_samples > sample_radius:
                    raise RuntimeError(
                        f"sample radius: {sample_radius} cannot be less than the number of requested samples: {num_samples}"
                    )

    def _init(self, dataset):
        assert isinstance(dataset, BaseEpisodesData), f"data ({type(dataset)}) is not an instance of {BaseEpisodesData}"
        # TODO: reference to dataset is not ideal here
        self._dataset = dataset

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Sampling                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _sample_idx(self, idx):
        # TODO: are we actually sampling distances correctly?
        # sample for observations
        episode, idx, offset = self._dataset.get_episode_and_idx(idx)
        indices = self.sample_episode_indices(episode, idx, n=self._num_samples, radius=self._sample_radius)
        # transform back to original indices
        return tuple(i + offset for i in indices)

    def sample_episode_indices(self, episode, idx, n=1, radius=None):
        # TODO: update this to use the same API
        #       as ground truth triplet and pair.
        # default value
        if radius is None:
            radius = len(episode)
        elif radius < 0:
            # TODO: I think this is the same as None? radius > len(episode) makes no difference?
            radius = len(episode) + radius + 1
        assert n <= len(episode)
        assert n <= radius
        # sample values
        attempts, indices = 0, {idx}
        while (len(indices) < n) and (attempts < n * 100):
            indices.add(sample_radius_fn(idx, low=0, high=len(episode), r_low=0, r_high=radius))
            attempts += 1
        # checks
        if len(indices) != n:
            raise RuntimeError("consider increasing the radius")
        # sort indices from highest to lowest.
        # - anchor is the newest
        # - positive is close in the past
        # - negative is far in the past
        return sorted(indices)[::-1]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from disent.nn.functional._conv2d import torch_conv2d_channel_wise
from disent.nn.functional._conv2d import torch_conv2d_channel_wise_fft

from disent.nn.functional._conv2d_kernels import get_kernel_size
from disent.nn.functional._conv2d_kernels import torch_gaussian_kernel
from disent.nn.functional._conv2d_kernels import torch_gaussian_kernel_2d
from disent.nn.functional._conv2d_kernels import torch_box_kernel
from disent.nn.functional._conv2d_kernels import torch_box_kernel_2d

from disent.nn.functional._correlation import torch_cov_matrix
from disent.nn.functional._correlation import torch_corr_matrix
from disent.nn.functional._correlation import torch_rank_corr_matrix
from disent.nn.functional._correlation import torch_pearsons_corr_matrix
from disent.nn.functional._correlation import torch_spearmans_corr_matrix

from disent.nn.functional._dct import torch_dct
from disent.nn.functional._dct import torch_idct
from disent.nn.functional._dct import torch_dct2
from disent.nn.functional._dct import torch_idct2

from disent.nn.functional._mean import torch_mean_generalized
from disent.nn.functional._mean import torch_mean_quadratic
from disent.nn.functional._mean import torch_mean_geometric
from disent.nn.functional._mean import torch_mean_harmonic

from disent.nn.functional._norm import torch_norm
from disent.nn.functional._norm import torch_dist
from disent.nn.functional._norm import torch_norm_euclidean
from disent.nn.functional._norm import torch_norm_manhattan
from disent.nn.functional._norm import torch_dist_hamming

from disent.nn.functional._other import torch_normalize
from disent.nn.functional._other import torch_nan_to_num
from disent.nn.functional._other import torch_unsqueeze_l
from disent.nn.functional._other import torch_unsqueeze_r

from disent.nn.functional._pca import torch_pca_eig
from disent.nn.functional._pca import torch_pca_svd
from disent.nn.functional._pca import torch_pca

# from disent.nn.functional._util_generic import TypeGenericTensor
# from disent.nn.functional._util_generic import TypeGenericTorch
# from disent.nn.functional._util_generic import TypeGenericNumpy
# from disent.nn.functional._util_generic import generic_as_int32
# from disent.nn.functional._util_generic import generic_max
# from disent.nn.functional._util_generic import generic_min
# from disent.nn.functional._util_generic import generic_shape
# from disent.nn.functional._util_generic import generic_ndim



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import math
from typing import Union

import numpy as np

# ========================================================================= #
# linear interpolate                                                        #
# ========================================================================= #


def lerp(ratio, start_val, end_val):
    """Linear interpolation between parameters, respects bounds when t is out of bounds [0, 1]"""
    # assert a < b
    r = np.clip(ratio, 0.0, 1.0)
    # precise method, guarantees v==b when t==1 | simplifies to: a + t*(b-a)
    return (1 - r) * start_val + r * end_val
    # return start_val + r * (end_val - start_val)  # EQUIVALENT


def lerp_step(step, max_step, start_val, end_val):
    """Linear interpolation based on a step count."""
    assert max_step > 0
    return lerp(ratio=step / max_step, start_val=start_val, end_val=end_val)


# ========================================================================= #
# linear interpolate                                                        #
# ========================================================================= #


_SCALE_RATIO_FNS = {
    "linear": lambda r: r,
    "sigmoid": lambda r: 1 / (1 + np.exp(-12 * r + 6)),
    "cosine": lambda r: 0.5 * (1 - np.cos(r * math.pi)),
}


def scale_ratio(r, mode="linear"):
    r = np.clip(r, 0.0, 1.0)
    return _SCALE_RATIO_FNS[mode](r)


# ========================================================================= #
# Cyclical Annealing Schedules                                              #
# - https://arxiv.org/abs/1903.10145                                        #
# - https://github.com/haofuml/cyclical_annealing                           #
# These functions are not exactly the same, but are more flexible.          #
# ========================================================================= #


_END_VALUES = {
    "low": 0,
    "high": 1,
}


def cyclical_anneal(
    step: Union[int, float, np.ndarray],
    period: float = 3600,
    low_ratio: float = 0.0,
    high_ratio: float = 0.0,
    repeats: int = None,
    start_low: bool = True,
    end_value: str = "high",
    mode: str = "linear",
):
    # check values
    assert 0 <= low_ratio <= 1
    assert 0 <= high_ratio <= 1
    assert (low_ratio + high_ratio) <= 1
    assert period > 0
    # compute ratio & masks
    r = (step / period) % 1
    # flip the axis
    if not start_low:
        r = 1 - r
    # get the clip mask
    low_mask, high_mask = r <= low_ratio, (1 - high_ratio) <= r
    # compute increasing values
    if low_ratio + high_ratio < 1:
        r = (r - low_ratio) / (1 - low_ratio - high_ratio)
        r = scale_ratio(r, mode=mode)
    # truncate values
    r = np.where(low_mask, 0, r)
    r = np.where(high_mask, 1, r)
    # repeats
    if repeats is not None:
        end_value = _END_VALUES.get(end_value, end_value)
        assert 0 <= end_value <= 1
        assert repeats > 0
        # compute
        n = step / period
        r = np.where(n < repeats, r, end_value)
    # done
    return r


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from typing import Sequence

import torch
from torch.distributions import Normal

from disent.frameworks.helper.util import compute_ave_loss_and_logs
from disent.frameworks.vae._unsupervised__betavae import BetaVae

# ========================================================================= #
# Beta-TC-VAE                                                               #
# ========================================================================= #


class BetaTcVae(BetaVae):
    """
    β-TCVAE - Isolating Sources of Disentanglement in VAEs
    https://arxiv.org/abs/1802.04942

    TODO: verify
    TODO: loss scales are not correct!
    TODO: simplify, some code is duplicated!
    Reference implementation is from: https://github.com/amir-abdi/disentanglement-pytorch
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(BetaVae.cfg):
        """
        Equation (4) with alpha=gamma=1 can be written as ELBO+(1-beta)*TC
        """

    # --------------------------------------------------------------------- #
    # Overrides                                                             #
    # --------------------------------------------------------------------- #

    def compute_ave_reg_loss(self, ds_posterior: Sequence[Normal], ds_prior: Sequence[Normal], zs_sampled):
        # compute kl loss
        # TODO: this should be KL instead? not KL Reg?
        kl_reg_loss, logs_kl_reg = super().compute_ave_reg_loss(ds_posterior, ds_prior, zs_sampled)
        # compute dip loss
        tc_reg_loss, logs_tc_reg = compute_ave_loss_and_logs(self._betatc_compute_loss, ds_posterior, zs_sampled)
        # combine
        combined_loss = kl_reg_loss + tc_reg_loss
        # return logs
        return combined_loss, {
            **logs_kl_reg,  # kl_loss, kl_reg_loss
            **logs_tc_reg,  #
        }

    # --------------------------------------------------------------------- #
    # Helper                                                                #
    # --------------------------------------------------------------------- #

    def _betatc_compute_loss(self, d_posterior: Normal, z_sampled):
        tc_loss = BetaTcVae._betatc_compute_total_correlation(
            z_sampled=z_sampled,
            z_mean=d_posterior.mean,
            z_logvar=torch.log(d_posterior.variance),
        )
        tc_reg_loss = (self.cfg.beta - 1.0) * tc_loss
        return tc_reg_loss, {
            "tc_loss": tc_loss,
            "tc_reg_loss": tc_reg_loss,
        }

    @staticmethod
    def _betatc_compute_total_correlation(z_sampled, z_mean, z_logvar):
        """
        Estimate total correlation over a batch.
        Reference implementation is from: https://github.com/YannDubs/disentangling-vae
        """
        # Compute log(q(z(x_j)|x_i)) for every sample in the batch, which is a
        # tensor of size [batch_size, batch_size, num_latents]. In the following
        # comments, [batch_size, batch_size, num_latents] are indexed by [j, i, l].
        log_qz_prob = BetaTcVae._betatc_compute_gaussian_log_density(
            z_sampled.unsqueeze(dim=1), z_mean.unsqueeze(dim=0), z_logvar.unsqueeze(dim=0)
        )

        # Compute log prod_l p(z(x_j)_l) = sum_l(log(sum_i(q(z(z_j)_l|x_i)))
        # + constant) for each sample in the batch, which is a vector of size
        # [batch_size,].
        log_qz_product = torch.logsumexp(log_qz_prob, dim=1, keepdim=False).sum(dim=1, keepdim=False)

        # Compute log(q(z(x_j))) as log(sum_i(q(z(x_j)|x_i))) + constant =
        # log(sum_i(prod_l q(z(x_j)_l|x_i))) + constant.
        log_qz = torch.logsumexp(log_qz_prob.sum(dim=2), dim=1, keepdim=False)

        return (log_qz - log_qz_product).mean()

    @staticmethod
    def _betatc_compute_gaussian_log_density(samples, mean, log_var):
        """
        Estimate the log density of a Gaussian distribution
        Reference implementation is from: https://github.com/amir-abdi/disentanglement-pytorch
        """
        # TODO: can this be replaced with some variant of Normal.log_prob?
        import math

        pi = torch.tensor(math.pi, requires_grad=False)
        normalization = torch.log(2 * pi)
        inv_sigma = torch.exp(-log_var)
        tmp = samples - mean
        return -0.5 * (tmp * tmp * inv_sigma + log_var + normalization)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import h5py
from torch.utils.data import Dataset

from disent.util.iters import LengthIter

# ========================================================================= #
# array utils                                                               #
# ========================================================================= #


class ArrayDataset(Dataset, LengthIter):
    def __init__(self, array, transform=None):
        self._array = array
        self._transform = transform

    def __len__(self):
        return self._array.shape[0]

    def __getitem__(self, item):
        assert isinstance(item, int)  # disable smart array accesses
        elem = self._array[item]
        if self._transform is not None:
            elem = self._transform(elem)
        return elem

    @property
    def shape(self):
        return self._array.shape


# ========================================================================= #
# hdf5 pickle dataset                                                       #
# ========================================================================= #


class Hdf5Dataset(Dataset, LengthIter):
    """
    This class supports pickling and unpickling of a read-only
    SWMR h5py file and corresponding dataset.

    WARNING: this should probably not be used across multiple hosts...
    """

    def __init__(self, h5_path: str, h5_dataset_name: str = "data", transform=None):
        self._h5_path = h5_path
        self._h5_dataset_name = h5_dataset_name
        self._hdf5_file, self._hdf5_data = self._make_hdf5()
        self._transform = transform

    def _make_hdf5(self):
        # TODO: can this cause a memory leak if it is never closed?
        hdf5_file = h5py.File(self._h5_path, "r", swmr=True)
        hdf5_data = hdf5_file[self._h5_dataset_name]
        return hdf5_file, hdf5_data

    def __len__(self):
        return self._hdf5_data.shape[0]

    def __getitem__(self, item):
        elem = self._hdf5_data[item]
        if self._transform is not None:
            elem = self._transform(elem)
        return elem

    @property
    def shape(self):
        return self._hdf5_data.shape

    def numpy_dataset(self) -> ArrayDataset:
        # TODO: make this function global
        return ArrayDataset(array=self._hdf5_data[:], transform=self._transform)

    def __enter__(self):
        return self

    def __exit__(self, error_type, error, traceback):
        self.close()

    # CUSTOM PICKLE HANDLING -- h5py files are not supported!
    # https://docs.python.org/3/library/pickle.html#pickle-state
    # https://docs.python.org/3/library/pickle.html#object.__getstate__
    # https://docs.python.org/3/library/pickle.html#object.__setstate__
    # TODO: this might duplicate in-memory stuffs.

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("_hdf5_file", None)
        state.pop("_hdf5_data", None)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._hdf5_file, self._hdf5_data = self._make_hdf5()

    def close(self):
        self._hdf5_file.close()
        del self._hdf5_file
        del self._hdf5_data

    def get_attrs(self) -> dict:
        return dict(self._hdf5_data.attrs)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import numpy as np

# ========================================================================= #
# Better Choice                                                             #
# ========================================================================= #


def random_choice_prng(a, size=None, replace=True, seed: int = None):
    # generate a random seed
    if seed is None:
        seed = np.random.randint(0, 2**32)
    # create seeded pseudo random number generator
    # - built in np.random.choice cannot handle large values: https://github.com/numpy/numpy/issues/5299#issuecomment-497915672
    # - PCG64 is the default: https://numpy.org/doc/stable/reference/random/bit_generators/index.html
    # - PCG64 has good statistical properties and is fast: https://numpy.org/doc/stable/reference/random/performance.html
    g = np.random.Generator(np.random.PCG64(seed=seed))
    # sample indices
    choices = g.choice(a, size=size, replace=replace)
    # done!
    return choices


# ========================================================================= #
# Random Ranges                                                             #
# ========================================================================= #


def randint2(a_low, a_high, b_low, b_high, size=None):
    """
    Like np.random.randint, but supports two ranges of values.
    Samples with equal probability from both ranges.
    - a: [a_low, a_high) -> including a_low, excluding a_high!
    - b: [b_low, b_high) -> including b_low, excluding b_high!
    """
    # convert
    a_low, a_high = np.array(a_low), np.array(a_high)
    b_low, b_high = np.array(b_low), np.array(b_high)
    # checks
    assert np.all(a_low <= a_high), f"a_low <= a_high | {a_low} <= {a_high}"
    assert np.all(b_low <= b_high), f"b_low <= b_high | {b_low} <= {b_high}"
    assert np.all(a_high <= b_low), f"a_high <= b_low | {a_high} <= {b_low}"
    # compute
    da = a_high - a_low
    db = b_high - b_low
    d = da + db
    assert np.all(
        d > 0
    ), f"(a_high - a_low) + (b_high - b_low) > 0 | {d} = ({a_high} - {a_low}) + ({b_high} - {b_low}) > 0"
    # sampled
    offset = np.random.randint(0, d, size=size)
    offset += (da <= offset) * (b_low - a_high)
    return a_low + offset


def sample_radius(value, low, high, r_low, r_high):
    """
    Sample around the given value (low <= value < high),
    the resampled value will lie in th same range.
    - sampling occurs in a radius around the value
      r_low <= radius < r_high
    """
    value = np.array(value)
    assert np.all(low <= value)
    assert np.all(value < high)
    # sample for new value
    return randint2(
        a_low=np.maximum(value - r_high + 1, low),
        a_high=np.maximum(value - r_low + 1, low),
        # if r_min == 0, then the ranges overlap, so we must shift one of them.
        b_low=np.minimum(value + r_low + (r_low == 0), high),
        b_high=np.minimum(value + r_high, high),
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
